==> 0.6.1&file=out_yasapi-0.6.1-src.tar.gz: out_yasapi-0.6.1/out_yasapi/yasapi_player_strategy.c <==
/*
 * yasapi_player_strategy.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>
#include <avrt.h>

///////////////////////////////////////////////////////////////////////////////
//#define YASAPI_SRC_DEFAULT_QUALITY

///////////////////////////////////////////////////////////////////////////////
#if ! defined (AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY) // {
#define AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY 0x08000000
#endif // }

#if ! defined(AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM) // {
#define AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM 0x80000000
#endif // }

///////////////////////////////////////////////////////////////////////////////
static HRESULT PlayerInitializeAudioClientPush(Player *pPlayer,
    REFERENCE_TIME hnsDevicePeriod)
{
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  AUDCLNT_SHAREMODE eShareMode=pPlayer->pPlugin->pOptions->eShareMode;
  DWORD dwStreamFlags=0;
#if defined (YASAPI_RESAMPLE) // {
  const WAVEFORMATEX *pwfx=pPlayer->write->Format(pPlayer);
#else // } {
  const WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
#endif // }

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode) {
#if defined (YASAPI_SRC_DEFAULT_QUALITY) // {
  	dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
#endif // }
  	dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
  }

  return pAudioClient->lpVtbl->Initialize(pAudioClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    0,                    // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static int PlayerCreateEventPush(Player *pPlayer)
{
  pPlayer->hEvent=NULL;

  return 0;
}

static void PlayerDestroyEventPush(Player *pPlayer)
{
}

static BOOL PlayerCanReadPush(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;

  if (0==pPlayer->ring.dwWritten)
    DPUTS(debug,"Warning: underflow\n");
#endif // }

  return 0<pPlayer->ring.dwWritten;
}

HRESULT PlayerGetFramesWritePush(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  UINT32 nFramesPadding;
  UINT32 nFramesWrite;
  UINT32 nFramesRing;
  HRESULT hr;

  // get padding.
  hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
    &nFramesPadding       // [out]  UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting padding");
    goto padding;
  }

  // caclulate the number of frames to render.
  nFramesWrite=nFrames-nFramesPadding;
  nFramesRing=pRing->dwWritten/pwfx->nBlockAlign;

  if (nFramesWrite>nFramesRing)
    nFramesWrite=nFramesRing;

  *pNumFramesWrite=nFramesWrite;
  *pNumFramesPadding=nFramesPadding;
padding:
  return hr;
}

static int PlayerWaitPush(Player *pPlayer, LONGLONG time)
{
  TimerSleep(&pPlayer->pPlugin->timer,time);

  return 0;
}

const PlayerStrategy *PlayerPush(void)
{
  static int initialized;
  static PlayerStrategy strategy;

  if (0==initialized) {
    strategy.name="push";
    strategy.InitializeAudioClient=PlayerInitializeAudioClientPush;
    strategy.CreateEvent=PlayerCreateEventPush;
    strategy.DestroyEvent=PlayerDestroyEventPush;
    strategy.CanRead=PlayerCanReadPush;
    strategy.Wait=PlayerWaitPush;
    strategy.GetFramesWrite=PlayerGetFramesWritePush;
    initialized=1;
  }

  return &strategy;
}

///////////////////////////////////////////////////////////////////////////////
static HRESULT PlayerInitializeAudioClientPull(Player *pPlayer,
    REFERENCE_TIME hnsDevicePeriod)
{
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  AUDCLNT_SHAREMODE eShareMode=pPlayer->pPlugin->pOptions->eShareMode;
  DWORD dwStreamFlags=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;
#if defined (YASAPI_RESAMPLE) // {
  const WAVEFORMATEX *pwfx=pPlayer->write->Format(pPlayer);
#else // } {
  const WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
#endif // }

  if (AUDCLNT_SHAREMODE_SHARED==eShareMode) {
#if defined (YASAPI_SRC_DEFAULT_QUALITY) // {
  	dwStreamFlags|=AUDCLNT_STREAMFLAGS_SRC_DEFAULT_QUALITY;
#endif // }
  	dwStreamFlags|=AUDCLNT_STREAMFLAGS_AUTOCONVERTPCM;
  }

  return pAudioClient->lpVtbl->Initialize(pAudioClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static int PlayerCreateEventPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  // create event.
  pPlayer->hEvent=CreateEvent(
    NULL,         // _In_opt_  LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,        // _In_      BOOL bManualReset,
    FALSE,        // _In_      BOOL bInitialState,
    NULL          // _In_opt_  LPCTSTR lpName
  );

  if (NULL==pPlayer->hEvent) {
    DMESSAGE(debug,"creating event");
    goto event;
  }

  DPUTS(debug,"  event created\n");

  // set event.
  hr=pAudioClient->lpVtbl->SetEventHandle(pAudioClient,
    pPlayer->hEvent           // [in]  HANDLE eventHandle
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
    DMESSAGE(debug,"setting event to the audio client");
    goto set;
  }

  DPUTS(debug,"  event set to the audio client\n");

  return 0;
set:
  CloseHandle(pPlayer->hEvent);
event:
  return -1;
}

static void PlayerDestroyEventPull(Player *pPlayer)
{
  CloseHandle(pPlayer->hEvent);
}

#if 0 // {
static BOOL PlayerCanReadPull(Player *pPlayer)
{
  return pPlayer->nFrames*pPlayer->wfxx.Format.nBlockAlign
      <=pPlayer->ring.dwWritten;
}
#else // } {
static BOOL PlayerCanReadPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;

  if (0==pPlayer->ring.dwWritten)
    DPUTS(debug,"Warning: underflow\n");
#endif // }

  return 0<pPlayer->ring.dwWritten;
}
#endif // }

static int PlayerWaitPull(Player *pPlayer, LONGLONG time)
{
  return WAIT_OBJECT_0==WaitForSingleObject(pPlayer->hEvent,2000)?0:-1;
}

HRESULT PlayerGetFramesWritePull(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding)
{
  *pNumFramesWrite=pPlayer->nFrames;
  *pNumFramesPadding=0;

  return S_OK;
}

const PlayerStrategy *PlayerPull(void)
{
  static int initialized;
  static PlayerStrategy strategy;

  if (0==initialized) {
    strategy.name="pull";
    strategy.InitializeAudioClient=PlayerInitializeAudioClientPull;
    strategy.CreateEvent=PlayerCreateEventPull;
    strategy.DestroyEvent=PlayerDestroyEventPull;
    strategy.CanRead=PlayerCanReadPull;
    strategy.Wait=PlayerWaitPull;
#if 0 // {
    strategy.GetFramesWrite=PlayerGetFramesWritePull;
#else // } {
    strategy.GetFramesWrite=PlayerGetFramesWritePush;
#endif // }
    initialized=1;
  }

  return &strategy;
}

==> 0.8.0&file=out_yasapi-0.8.0-src.tar.gz: out_yasapi-0.8.0/out_yasapi/yasapi_player_strategy.c <==
/*
 * yasapi_player_strategy.c
 * Copyright (C) 2015 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static HRESULT PlayerInitializeAudioClientPush(Player *pPlayer,
    DWORD dwStreamFlags, REFERENCE_TIME hnsDevicePeriod)
{
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Options *pOptions=pPlayer->pPlugin->pOptions;
  AUDCLNT_SHAREMODE eShareMode=OptionsGetShareMode(pOptions);
  const WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;

  return pAudioClient->lpVtbl->Initialize(pAudioClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    0,                    // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static int PlayerCreateEventPush(Player *pPlayer)
{
  pPlayer->hEvent=NULL;

  return 0;
}

static void PlayerDestroyEventPush(Player *pPlayer)
{
}

static BOOL PlayerCanReadPush(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;

  if (0==pPlayer->ring.dwWritten)
    DPUTS(debug,"Warning: underflow\n");
#endif // }

  return 0<pPlayer->ring.dwWritten;
}

HRESULT PlayerGetFramesWritePush(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Ring *pRing=&pPlayer->ring;
  WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;
  UINT32 nFrames=pPlayer->nFrames;
  UINT32 nFramesPadding;
  UINT32 nFramesWrite;
  UINT32 nFramesRing;
  HRESULT hr;

  // get padding.
  hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
    &nFramesPadding       // [out]  UINT32 *pNumPaddingFrames
  );

  if (FAILED(hr)) {
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DERROR(debug,E_POINTER,hr);
    DMESSAGE(debug,"getting padding");
    goto padding;
  }

  // caclulate the number of frames to render.
  nFramesWrite=nFrames-nFramesPadding;
  nFramesRing=pRing->dwWritten/pwfx->nBlockAlign;

  if (nFramesWrite>nFramesRing)
    nFramesWrite=nFramesRing;

  *pNumFramesWrite=nFramesWrite;
  *pNumFramesPadding=nFramesPadding;
padding:
  return hr;
}

static int PlayerWaitPush(Player *pPlayer, LONGLONG time)
{
  TimerSleep(&pPlayer->pPlugin->timer,time);

  return 0;
}

const PlayerStrategy *PlayerPush(void)
{
  static int initialized;
  static PlayerStrategy strategy;

  if (0==initialized) {
    strategy.name="push";
    strategy.InitializeAudioClient=PlayerInitializeAudioClientPush;
    strategy.CreateEvent=PlayerCreateEventPush;
    strategy.DestroyEvent=PlayerDestroyEventPush;
    strategy.CanRead=PlayerCanReadPush;
    strategy.Wait=PlayerWaitPush;
    strategy.GetFramesWrite=PlayerGetFramesWritePush;
    initialized=1;
  }

  return &strategy;
}

///////////////////////////////////////////////////////////////////////////////
static HRESULT PlayerInitializeAudioClientPull(Player *pPlayer,
    DWORD dwStreamFlags, REFERENCE_TIME hnsDevicePeriod)
{
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  Options *pOptions=pPlayer->pPlugin->pOptions;
  AUDCLNT_SHAREMODE eShareMode=OptionsGetShareMode(pOptions);
  const WAVEFORMATEX *pwfx=&pPlayer->wfxx.Format;

  dwStreamFlags|=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;

  return pAudioClient->lpVtbl->Initialize(pAudioClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioSessionGuid
  );
}

static int PlayerCreateEventPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;
#else // } {
  int debug=0;
#endif // }
  IAudioClient *pAudioClient=pPlayer->pAudioClient;
  HRESULT hr;

  // create event.
  pPlayer->hEvent=CreateEvent(
    NULL,         // _In_opt_  LPSECURITY_ATTRIBUTES lpEventAttributes,
    FALSE,        // _In_      BOOL bManualReset,
    FALSE,        // _In_      BOOL bInitialState,
    NULL          // _In_opt_  LPCTSTR lpName
  );

  if (NULL==pPlayer->hEvent) {
    DMESSAGE(debug,"creating event");
    goto event;
  }

  DPUTS(debug,"  event created\n");

  // set event.
  hr=pAudioClient->lpVtbl->SetEventHandle(pAudioClient,
    pPlayer->hEvent           // [in]  HANDLE eventHandle
  );

  if (FAILED(hr)) {
    DERROR(debug,E_INVALIDARG,hr);
    DERROR(debug,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED,hr);
    DERROR(debug,AUDCLNT_E_NOT_INITIALIZED,hr);
    DERROR(debug,AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(debug,AUDCLNT_E_SERVICE_NOT_RUNNING ,hr);
    DMESSAGE(debug,"setting event to the audio client");
    goto set;
  }

  DPUTS(debug,"  event set to the audio client\n");

  return 0;
set:
  CloseHandle(pPlayer->hEvent);
event:
  return -1;
}

static void PlayerDestroyEventPull(Player *pPlayer)
{
  CloseHandle(pPlayer->hEvent);
}

#if 0 // {
static BOOL PlayerCanReadPull(Player *pPlayer)
{
  return pPlayer->nFrames*pPlayer->wfxx.Format.nBlockAlign
      <=pPlayer->ring.dwWritten;
}
#else // } {
static BOOL PlayerCanReadPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->pPlugin->pOptions->debug;

  if (0==pPlayer->ring.dwWritten)
    DPUTS(debug,"Warning: underflow\n");
#endif // }

  return 0<pPlayer->ring.dwWritten;
}
#endif // }

static int PlayerWaitPull(Player *pPlayer, LONGLONG time)
{
  return WAIT_OBJECT_0==WaitForSingleObject(pPlayer->hEvent,2000)?0:-1;
}

HRESULT PlayerGetFramesWritePull(Player *pPlayer, UINT32 *pNumFramesWrite,
    UINT32 *pNumFramesPadding)
{
  *pNumFramesWrite=pPlayer->nFrames;
  *pNumFramesPadding=0;

  return S_OK;
}

const PlayerStrategy *PlayerPull(void)
{
  static int initialized;
  static PlayerStrategy strategy;

  if (0==initialized) {
    strategy.name="pull";
    strategy.InitializeAudioClient=PlayerInitializeAudioClientPull;
    strategy.CreateEvent=PlayerCreateEventPull;
    strategy.DestroyEvent=PlayerDestroyEventPull;
    strategy.CanRead=PlayerCanReadPull;
    strategy.Wait=PlayerWaitPull;
#if 0 // {
    strategy.GetFramesWrite=PlayerGetFramesWritePull;
#else // } {
    strategy.GetFramesWrite=PlayerGetFramesWritePush;
#endif // }
    initialized=1;
  }

  return &strategy;
}
